"""A python dice library."""
from .dice_logic import Roll
from .elements import Die, DiceGroup
import logging

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger('snakeeyes')
logger.setLevel(logging.DEBUG)
handler = logging.FileHandler(filename='snakeeyes.log', encoding='utf-8', mode='w')
handler.setFormatter(logging.Formatter('%(asctime)s:%(levelname)s:%(name)s: %(message)s'))
logger.addHandler(handler)



import snakeeyes


class TestDieBools():
    @staticmethod
    def test_die():
        """
        Set the test test.

        Args:
            self: (todo): write your description
        """
        die = bool(snakeeyes.DiceGroup("28*15"))
        assert die is False



"""
Handles the actual logic of dice rolls

Classes
-------
Roll

Functions
---------
roll
"""
import math
import random
import re
import logging
from .elements import DiceGroup, Exploding, Successes, Die

import ast
import operator

binOps = {
    ast.Add: operator.add,
    ast.Sub: operator.sub,
    ast.Mult: operator.mul,
    ast.Div: operator.truediv,
    ast.Mod: operator.mod
}


def arithmeticEval(s):
    """
    Evaluate an operator.

    Args:
        s: (todo): write your description
    """
    node = ast.parse(s, mode='eval')

    def _eval(node):
        """
        Evaluate an ast node.

        Args:
            node: (todo): write your description
        """
        if isinstance(node, ast.Expression):
            return _eval(node.body)
        if isinstance(node, ast.Str):
            return node.s
        if isinstance(node, ast.Num):
            return node.n
        if isinstance(node, ast.BinOp):
            return binOps[type(node.op)](_eval(node.left), _eval(node.right))
        else:
            raise Exception('Unsupported type {}'.format(node))

    return _eval(node.body)


logger = logging.getLogger('snakeeyes.dicelogic')


op_dict = {
    ">": Successes,
    "x": Exploding,
}


def roll(die: Die):
    """
    Takes Die object and returns a tuple containing a list of results, and a total of of all rolls.

    Parameters
    ----------

    die : elements.Die
    """
    if die:
        dice_array = []
        for i in range(die.quantity):
            dice_array.append(math.ceil(random.random() * die.sides))
        dice_total = 0
        for r in dice_array:
            dice_total += r
        return (dice_array, dice_total)
    else:
        return False


class Roll():

    """A class which takes a string and outputs a dice roll

    Parameters
    ----------
    string : str
        The input string

    Attributes
    ----------
    dicer_regex : str
        Regex showing how to extract just the dice string from an object
    op_dict : dict
        dictionary containing the characters of an operator, and the assosciated class
    die : elements.Die
        Dice roll using string input
    results : list of int
        List of results from dice rolled
    total : int
        The total of all dice rolled

    Methods
    -------
    op_collection(die)
        Return operator class list
    op_evaluate(ops)
        Return the final result of all operators

    """

    dice_regex = re.compile(r"\d*d\d*(?:[^d\d\(\)+\-\*/]\d*)*")
    math_regex = re.compile(r"[\(\)+*-\/\d]+")

    @staticmethod
    def op_collection(die: Die):
        """
        Take die object and return list of operator classes.

        Parameters
        ----------
        die : elements.Die

        Returns
        -------
        ops : list of (elements.operator, int)
        """
        oper = []
        for o in die.ops:
            try:
                operator = op_dict[o[0]]
                op = (operator, o[1])
                oper.append(op)
            except KeyError:
                continue
        oper = sorted(oper, key=lambda op: op[0].priority)
        return oper

    @staticmethod
    def op_evaluate(die: Die, ops: list, results: list):
        """Take results and operators and return a final result."""
        last_output = results
        last_output = ops[0].evaluate(last_output, ops[1], die)
        return last_output

    def __init__(self, string: str):
        """
        Init results from the input string.

        Args:
            self: (todo): write your description
            string: (todo): write your description
        """
        self.string = string
        logger.debug("Trying!")
        self.die = DiceGroup(self.string)
        self.rolls = []
        self.total = 0
        self.results = []
        self.successes = False
        if self.die.dice:
            # If there are actual dice in the roll, then do this
            for d in self.die.dice:
                r = roll(d)
                self.rolls.append((r, d.string, d))
            for r in self.rolls:
                # This is here to basically turn all the rolls into usable stuff and then output it
                # in a way that actually makes sense
                tempstring = re.compile(rf"{r[1]}")
                if r[0]:
                    tempresults = r[0][0]
                    self.string = tempstring.sub(f"{r[0][1]}", self.string)
                    self.total += r[0][1]
                    if r[2].ops:
                        for o in self.op_collection(r[2]):
                            tempresults = self.op_evaluate(r[2], o, tempresults)
                            if o[0] is Successes:
                                self.successes = True
                                break
                    self.results.append((r[1], tempresults))
        self.final = arithmeticEval(self.string)



"""
Handles the Grammar of the document

Classes
-------
- DiceString
- Die
- Operator
    - LeftHandOperator
        - Successes
        - Exploding

"""
import logging
import math
import random
import re

logger = logging.getLogger('snakeeyes.elements')


class Die():
    """
    Generates the dice string to put through the system

        Attributes
        ----------
        parsestring : Pattern
            Regex pattern to detect the various attributes of a dice roll

        quantity : int
            Number of times die must be rolled

        sides : int
            Number of sides a die has

    """
    sides = 0
    quantity = 0

    def __init__(self, quantity: int, sides: int, oplist: list, string: str):
        """
        Initialize the string.

        Args:
            self: (todo): The instance
            quantity: (int): Number of dice to be rolled
            sides: (int): String which contains the sides of each die. Should be castable to int.
        """
        logger.debug("Initating DiceString")
        try:
            self.quantity = quantity
            self.sides = sides
            self.ops = oplist
            self.string = string
            logger.debug("Sides: %i \n Quantity: %i ",
                         self.sides, self.quantity)
        except (ValueError, AttributeError):
            pass


class DiceGroup():
    """Class that handles dice rolls using regular expressions.

    Attributes
    ----------
    string : str
        String to be processed
    dice : list
        List of dice rolled.
    """
    parsestring = re.compile(r"(?P<quantity> \d* (?=d\d*)) d (?P<sides>\d*)(?:[x\>]\d*)*", re.X)
    parseops = re.compile(r"(?P<operator>[x\>]) (?P<operands>\d*)", re.X)

    def __init__(self, string: str):
        """
        Initialize a group of dice from the string.

        Args:
            self: The Instance.
            string: (str): The string to be processed.
        """
        self.string = string
        logger.debug("String: %s", self.string)
        __dice = self.parsestring.finditer(string)
        self.dice = []
        for d in __dice:
            # Intiialize the dice
            quant = int(d.group('quantity'))
            sides = int(d.group('sides'))
            dstring = d.group()
            ops = self.parseops.finditer(dstring)
            opslist = []
            for o in ops:
                # initialize operator tuples
                optuple = (o.group('operator'), int(o.group('operands')))
                opslist.append(optuple)
            dicestring = Die(quant, sides, opslist, dstring)
            self.dice.append(dicestring)

            logger.debug(f"Oplist: {opslist}")
        logger.debug(f"Dice: {self.dice}")

    def __bool__(self):
        """
        Returns true if the dice is true false otherwise.

        Args:
            self:: The instance
        """
        if self.dice:
            logger.debug("Die is true")
            return True
        return False


class Operator():
    """Handles creating operators for use in rolls.

    ...

    Attributes
    ----------
    char : str
        character for operand
    regex : str
        raw string, by default just detects the character

    Functions
    -------
    parse - Take the string and output operator and operands
    evaluate - Blank method where the operator is processed

    """
    priority = 0
    char = r""

    @classmethod
    def evaluate(cls, results: list, operand: int, die: Die):
        """
        Evaluate the given dice.

        Args:
            cls: (callable): The class
            dice: (Die): The Die being rolled
        """


class LeftHandOperator(Operator):
    """Operators that act on the object to the left, using the object on the right, inherits from Operator.

    Attributes
    ----------
    operand : str
        The arguments taken by the operator

    """
    operand = r"\d*"


class Successes(LeftHandOperator):
    """Takes an operand and calculates how many successes there have been."""
    priority = 7
    char = r"\>"

    @classmethod
    def evaluate(cls, results: list, operand: int, die: Die):
        """
        Evaluate a list of results.

        Args:
            cls: (callable): The Class
            results: (list): List of Results.
            operand: (int): The threshold after which die count as successes.
            die: (Die): The Die roll.
        """
        dice_list = []
        logger.debug("Evaluating successes!")
        for d in results:
            if d > int(operand):
                dice_list.append((d, True))
            else:
                dice_list.append((d, False))

        return dice_list


class Exploding(LeftHandOperator):
    """
    Takes dice results, and if the value is greater than the threshold, rolls another die.
    """
    priority = 1
    char = r"x"

    @classmethod
    def evaluate(cls, results: list, operand: int, die: Die):
        """
        Evaluate the objective function.

        Args:
            cls: (callable): The class
            results: (list): The list of results
            operand: (int): The threshold at which the die is rerolled
            die: (Die): The Die being rolled
        """
        eval_results = results
        logger.debug("Evaluating Exploding!")
        for d in results:
            logger.debug("D is: %i", d)
            r = d
            logger.debug("Operand: %i", operand)
            while r >= operand:
                logger.debug("Exploded!")
                temp_roll = math.ceil(random.random() * die.sides)
                r = temp_roll
                logger.debug(" R is %i", r)
                eval_results.append(temp_roll)
                if r >= operand:
                    break
        logger.debug("Exploded dice: %s", str(eval_results))
        return eval_results



import snakeeyes


def test_roll():
    """
    Test whether the test

    Args:
    """
    return snakeeyes.Roll("1d20")


class TestMath():
    def test_add(self):
        """
        Returns the test test for the test.

        Args:
            self: (todo): write your description
        """
        return snakeeyes.Roll("1d20+5")

    def test_sub(self):
        """
        Returns the test sub - test test.

        Args:
            self: (todo): write your description
        """
        return snakeeyes.Roll("1d20-5")

    def test_mult(self):
        """
        Return the test test test test.

        Args:
            self: (todo): write your description
        """
        return snakeeyes.Roll("1d20*5")

    def test_div(self):
        """
        Return the test test.

        Args:
            self: (todo): write your description
        """
        return snakeeyes.Roll("1d20/5")

    def test_raw(self):
        """
        Returns raw test test test test.

        Args:
            self: (todo): write your description
        """
        return snakeeyes.Roll("25*3")

    def test_parenth(self):
        """
        Return the test test test.

        Args:
            self: (todo): write your description
        """
        return snakeeyes.Roll("(1d20+5)*20")

class TestMultiDice():
    @staticmethod
    def test_add():
        """
        Add test test to a test

        Args:
        """
        return snakeeyes.Roll("1d6+2d10")
    @staticmethod
    def test_subtract():
        """
        Subtract test test test test

        Args:
        """
        return snakeeyes.Roll("3d20-6d20")
    @staticmethod
    def test_divide():
        """
        Divide a test test.

        Args:
        """
        return snakeeyes.Roll("7d10/3d20")
    @staticmethod
    def test_mult():
        """
        Return a test test function

        Args:
        """
        return snakeeyes.Roll("3d20*2d10")
    @staticmethod
    def test_exp():
        """
        Return the test test.

        Args:
        """
        return snakeeyes.Roll("3d20x15*2d10x15")
    @staticmethod
    def test_succ():
        """
        Returns the test test suite.

        Args:
        """
        return snakeeyes.Roll("3d20>10+2d20>10")


class TestOperators():
    def test_success(self):
        """
        Return the test test.

        Args:
            self: (todo): write your description
        """
        return snakeeyes.Roll("1d20>5")

    def test_explode(self):
        """
        Return the test test.

        Args:
            self: (todo): write your description
        """
        return snakeeyes.Roll("100d20x10")
